# coding=utf-8
from __future__ import absolute_import, division, print_function

import os

from suanpan import g
from suanpan import path as spath
from suanpan import runtime
from suanpan.log import logger
from suanpan.storage import base
from suanpan.utils import pbar as spbar


class Storage(base.Storage):
    def __init__(
        self,
        localTempStore=base.Storage.DEFAULT_TEMP_DATA_STORE,
        localGlobalStore=base.Storage.DEFAULT_GLOBAL_DATA_STORE,
        **kwargs,
    ):
        super().__init__(
            delimiter=os.sep,
            tempStore=localTempStore,
            globalStore=localGlobalStore,
            **kwargs,
        )

    def download(
        self, key, path=None, progress=None, quiet=not g.debug
    ):  # pylint: disable=unused-argument
        url = self.storageUrl(key)
        tmppath = self.getPathInTempStore(key)
        path = path or tmppath

        if not quiet:
            logger.debug(f"Downloading: {url} -> {path}")

        with spbar.one(
            total=1, disable=quiet, desc="Downloading", progress=progress,
        ) as pbar:
            if path == tmppath:
                pbar.update(1)
                pbar.set_description("Nothing to do")
            else:
                spath.copy(tmppath, path)
                pbar.update(1)
                pbar.set_description("Downloaded")
            return path

    def upload(
        self, key, path=None, progress=None, quiet=not g.debug
    ):  # pylint: disable=unused-argument
        url = self.storageUrl(key)
        tmppath = self.getPathInTempStore(key)
        path = path or tmppath

        if not quiet:
            logger.debug(f"Uploading: {path} -> {url}")

        with spbar.one(
            total=1, disable=quiet, desc="Uploading", progress=progress,
        ) as pbar:
            if path == tmppath:
                pbar.update(1)
                pbar.set_description("Nothing to do")
            else:
                spath.copy(path, tmppath)
                pbar.update(1)
                pbar.set_description("Uploaded")
            return path

    def copy(
        self, src, dist, progress=None, quiet=not g.debug
    ):  # pylint: disable=unused-argument
        srcurl = self.storageUrl(src)
        disturl = self.storageUrl(dist)
        srcpath = self.getPathInTempStore(src)
        distpath = self.getPathInTempStore(dist)

        if not quiet:
            logger.debug(f"Copying: {srcurl} -> {disturl}")

        with spbar.one(
            total=1, disable=quiet, desc="Copying", progress=progress,
        ) as pbar:
            spath.copy(srcpath, distpath)
            pbar.update(1)
            pbar.set_description("Copied")
            return dist

    def remove(
        self, key, progress=None, quiet=not g.debug
    ):  # pylint: disable=unused-argument
        url = self.storageUrl(key)
        path = self.getPathInTempStore(key)

        if not quiet:
            logger.debug(f"Removing: {url}")

        with spbar.one(
            total=1, disable=quiet, desc="Removing", progress=progress,
        ) as pbar:
            spath.remove(path)
            pbar.update(1)
            pbar.set_description("Removed")
            return key

    def walk(self, key):
        path = self.getPathInTempStore(key)
        for root, folders, files in runtime.saferun(os.walk, default=iter(()))(path):
            root = self.completePath(spath.replaceBasePath(path, key, root)[0])
            folders = [self.storagePathJoin(root, folder) for folder in folders]
            files = [self.storagePathJoin(root, file) for file in files]
            yield root, folders, files

    def listAll(self, key):
        path = self.getPathInTempStore(key)
        return (self.storagePathJoin(key, item) for item in os.listdir(path))

    def listFolders(self, key):
        return (item for item in self.listAll(key) if self.isFolder(item))

    def listFiles(self, key):
        return (item for item in self.listAll(key) if self.isFile(item))

    def isFolder(self, key):
        folder = self.getPathInTempStore(key)
        return os.path.isdir(folder)

    def isFile(self, key):
        file = self.getPathInTempStore(key)
        return os.path.isfile(file)

    def storagePathJoin(self, *paths):
        return self.localPathJoin(*paths)

    def storageRelativePath(self, path, basepath):
        return self.localRelativePath(path, basepath)

    def storageUrl(self, key):
        return "file://" + self.getPathInTempStore(key)

    def getStorageMd5(self, key):
        return self.getLocalMd5(self.getPathInTempStore(key))

    def getStorageSize(self, key):
        return self.getLocalSize(self.getPathInTempStore(key))



# coding=utf-8
from __future__ import absolute_import, print_function

from suanpan import asyncio
from suanpan.g import g
from suanpan.run import cli, helper, env, run
from suanpan.sfile import settings

__version__ = "0.19.8"
__all__ = ['asyncio', 'g', 'cli', 'helper', 'env', 'run', 'settings']

if __name__ == "__main__":
    print(f"Suanpan SDK (ver: {__version__})")



# coding=utf-8
from __future__ import absolute_import, print_function

import requests

from suanpan import error
from suanpan.api import auth as apiauth


def session(auth=True):
    sess = requests.Session()
    if auth:
        sess.headers.update(apiauth.defaultHeaders())
    return sess


def request(method, url, *args, **kwargs):
    sess = session(auth=kwargs.get("auth", True))
    func = getattr(sess, method)
    rep = func(url, *args, **kwargs)
    rep.raise_for_status()
    result = rep.json()
    if not result.get("success", True):
        raise error.ApiError(f"Request failed {result}")
    return result


def get(url, *args, **kwargs):
    return request("get", url, *args, **kwargs)


def post(url, *args, **kwargs):
    return request("post", url, *args, **kwargs)


def put(url, *args, **kwargs):
    return request("put", url, *args, **kwargs)


def delete(url, *args, **kwargs):
    return request("delete", url, *args, **kwargs)



# coding=utf-8
from __future__ import absolute_import, print_function

import os
from suanpan import error
from suanpan import path as spath
from suanpan.app import loops
from suanpan.app.base import BaseApp
from suanpan.app.modules import modules
from suanpan.model.arguments import HotReloadModel
from suanpan.stream import Stream
from suanpan.utils import functional, json

COMPONENT_ARGUMENT_CLASSES = (HotReloadModel,)


class StreamAppCore(object):
    def __call__(self, funcOrApp):
        if not isinstance(funcOrApp, StreamAppCore):
            self._handler.use(functional.instancemethod(funcOrApp))
        return self

    @property
    def name(self):
        return self._stream.name

    @functional.lazyproperty
    def _stream(self):
        raise NotImplementedError("Method not implemented!")

    @functional.lazyproperty
    def _loop(self):
        raise NotImplementedError("Method not implemented!")

    @functional.lazyproperty
    def _ergo_loop(self):
        raise NotImplementedError("Method not implemented!")

    @functional.lazyproperty
    def _handler(self):
        return self._loop.getHandler()

    def bootstrap(self):
        self.modules.info()

    @property
    def modules(self):
        return modules


class BaseStreamApp(StreamAppCore, BaseApp):
    def start(self, *args, **kwargs):
        self.bootstrap()
        super().start(*args, **kwargs)
        return self._stream.start(*args, **kwargs)

    def input(self, argument):
        if isinstance(argument, COMPONENT_ARGUMENT_CLASSES):
            self._stream.ARGUMENTS.append(argument)
        else:
            self._handler.input(argument)
        return self

    def output(self, argument):
        if isinstance(argument, HotReloadModel):
            raise error.AppError(f"{argument.name} can't be set as output!")
        self._handler.output(argument)
        return self

    def param(self, argument):
        self._stream.ARGUMENTS.append(argument)
        return self

    def column(self, argument):
        self._stream.ARGUMENTS.append(argument)
        return self

    def beforeInit(self, hook):
        self._stream.addBeforeInitHooks(hook)
        return hook

    def afterInit(self, hook):
        self._stream.addAfterInitHooks(hook)
        return hook

    def beforeCall(self, hook):
        self._handler.addBeforeCallHooks(hook)
        return hook

    def afterCall(self, hook):
        self._handler.addAfterCallHooks(hook)
        return hook

    def beforeExit(self, hook):
        self._stream.addBeforeExitHooks(hook)
        return hook

    def load(self, *args, **kwargs):
        return self._handler.load(*args, **kwargs)

    def send(self, *args, **kwargs):
        return self._loop.send(*args, **kwargs)

    def ergo_send(self, *args, **kwargs):
        return self._ergo_loop.ergo_send(*args, **kwargs)

    def save(self, *args, **kwargs):
        return self.send(*args, **kwargs)

    def publish(self, topic: str, payload: dict):
        data = json.dumps(payload)
        self._stream.pubsubLoop.send(topic, data)

    def subscribe(self, topic: str):
        def decorator(handler):
            self._stream.addEventHook(topic, handler)
            return handler

        return decorator

    def unsubscribe(self, topic: str):
        self._stream.deleteEventHook(topic)


class StreamApp(BaseStreamApp):
    @functional.lazyproperty
    def _stream(self):
        return Stream()

    @functional.lazyproperty
    def _loop(self):
        return self._stream.callLoop

    @functional.lazyproperty
    def _ergo_loop(self):
        return self._stream.ergoLoop

    @property
    def args(self):
        return self._stream.args

    @property
    def vars(self):
        return self._stream.vars

    @functional.lazyproperty
    def trigger(self):  # pylint: disable=invalid-overridden-method
        return TriggerApp(self._stream)

    @functional.lazyproperty
    def sio(self):
        return SIOAppManager(self._stream)

    def title(self, title):
        self._stream.title = title
        return self

    def bootstrap(self):
        super().bootstrap()
        self.trigger.bootstrap(self)
        self.sio.bootstrap(self)
        self.modules.bootstrap()


class TriggerApp(BaseStreamApp):
    def __init__(self, stream, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__stream = stream

    @functional.lazyproperty
    def _stream(self):
        return self.__stream

    @functional.lazyproperty
    def _loop(self):
        return self._stream.triggerLoop

    def interval(self, seconds, *args, **kwargs):
        self._loop.loop(loops.Interval(seconds, *args, **kwargs))
        return self

    def loop(self, funcOrIter, *args, **kwargs):
        self._loop.loop(funcOrIter, *args, **kwargs)
        return self

    def title(self, title):
        self._stream.title = title
        return self

    def bootstrap(self, app):
        pass


class SIOCore(StreamAppCore):
    def __init__(self, stream, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__stream = stream

    @functional.lazyproperty
    def _stream(self):
        return self.__stream

    @functional.lazyproperty
    def _loop(self):
        return self._stream.sioLoop

    def emit(self, *args, **kwargs):
        self._loop.emit(*args, **kwargs)


class SIOAppManagerCore(SIOCore):
    def __init__(self, stream, *args, **kwargs):
        super().__init__(stream, *args, **kwargs)
        self._siomap = {}

    def _handle(self, event, handlers):
        def _func(context):
            result = None
            for handler in handlers:
                result = handler(context)
            return result

        _func.__name__ = event
        return _func


class SIOAppCore(SIOCore):
    def __init__(self, stream, event, *args, **kwargs):
        super().__init__(stream, *args, **kwargs)
        self._event = event


class SIOAppManager(SIOAppManagerCore):
    def __init__(self, stream, *args, **kwargs):
        super().__init__(stream, *args, **kwargs)
        self._siomap = {}

    def __call__(self, event):
        if event not in self._siomap:
            self._siomap[event] = SIOApp(self._stream, event)
        return self._siomap[event]

    @property
    def contentTypes(self):
        return self._loop.contentTypes

    @property
    def staticFiles(self):
        return self._loop.extraStaticFiles

    def formatError(self, err):
        return self._loop.formatError(err)

    @functional.lazyproperty
    def stream(self):
        return SIOStreamAppManager(self._stream)

    def title(self, title):
        self._loop.options.params.configs.title = title
        return self

    def params(self, params):
        self._loop.options.params.configs.params = params
        return self

    def enter(self, *args, **kwargs):
        return self._loop.enter(*args, **kwargs)

    def leave(self, *args, **kwargs):
        return self._loop.leave(*args, **kwargs)

    def static(self, *paths):
        return os.path.join(self._loop.staticFolder, *paths)

    def setStatic(self, path, ignore=True):
        return spath.copy(functional.resourcepath(path), self._loop.staticFolder, ignore=ignore)

    def enable(self):
        self._loop.enable()
        return self

    def disable(self):
        self._loop.disable()
        return self

    def bootstrap(self, app):
        for event, handlers in app.modules.handlers.items():
            self(event)(self._handle(event, handlers))
        self.stream.bootstrap(app)


class SIOApp(SIOAppCore):
    @functional.lazyproperty
    def _handler(self):
        return self._loop.getHandler(self._event)


class SIOStreamAppManager(SIOAppManager):
    def __call__(self, event):
        if event not in self._siomap:
            self._siomap[event] = SIOStreamApp(self._stream, event)
        return self._siomap[event]

    def bootstrap(self, app):
        for event, handlers in app.modules.streamHandlers.items():
            self(event)(self._handle(event, handlers))


class SIOStreamApp(SIOAppCore):
    @functional.lazyproperty
    def _handler(self):
        return self._loop.siostream.getHandler(self._event)



# coding=utf-8
from __future__ import absolute_import, print_function

import gevent
import traceback

import redis
from collections import OrderedDict
from gevent.event import Event
from suanpan import error, utils, g
from suanpan.log import logger

MQ_DELIVER_KEY = "_suanpan_mq_deliver"


class MQ(object):
    def __init__(
            self,  # pylint: disable=unused-argument
            redisHost="localhost",
            redisPort=6379,
            redisRealtime=False,
            redisKeepalive=True,
            # redisKeepaliveIDLE=120,
            # redisKeepaliveCNT=2,
            # redisKeepaliveINTVL=30,
            redisSocketConnectTimeout=1,
            redisUnixSocketPath=None,
            options=None,
            client=None,
            __p2p_stream=False,
            **kwargs,
    ):
        self.options = options or {}
        self._p2p = __p2p_stream or kwargs.get('__p2p_stream', False)
        self.options.update(
            host=redisHost,
            port=redisPort,
            realtime=redisRealtime,
            keepalive=redisKeepalive,
            # keepidle=redisKeepaliveIDLE,
            # keepcnt=redisKeepaliveCNT,
            # keepintvl=redisKeepaliveINTVL,
            socketConnectTimeout=redisSocketConnectTimeout,
            unixSocketPath=redisUnixSocketPath,
        )
        if client and not isinstance(client, redis.Redis):
            raise error.MQError(f"Invalid redis client: {client}")

        if client:
            self.client = client
        elif self.options["unixSocketPath"]:
            self.client = redis.Redis(
                decode_responses=True, unix_socket_path=self.options["unixSocketPath"]
            )
        else:
            self.client = redis.Redis(
                host=self.options["host"],
                port=self.options["port"],
                decode_responses=True,
                socket_keepalive=self.options["keepalive"],
                # socket_keepalive_options={
                #     socket.TCP_KEEPIDLE: self.options["keepidle"],
                #     socket.TCP_KEEPCNT: self.options["keepcnt"],
                #     socket.TCP_KEEPINTVL: self.options["keepintvl"],
                # },
                socket_connect_timeout=self.options["socketConnectTimeout"],
            )

        self._stop_event = Event()
        self._stop_event.clear()

        if self._p2p:
            self._getP2PConnections()

    def close(self):
        self._stop_event.set()
        self._stop_event.is_set()
        try:
            self.client.close()
            self.client.connection_pool.disconnect()
        finally:
            pass

    @property
    def connected(self):
        return bool(self.client.connection)

    def _getQueue(self, name):
        try:
            return self.client.xinfo_stream(name)
        except redis.RedisError:
            return {}

    def _getQueueGroups(self, name):
        try:
            return self.client.xinfo_groups(name)
        except redis.RedisError:
            return []

    def _lenQueue(self, name):
        try:
            return self.client.xlen(name)
        except redis.RedisError:
            return 0

    def listQueueNames(self):
        return self.client.keys("mq-*")

    def getQueueInfo(self, name):
        # queue = self._getQueue(name)
        length = self._lenQueue(name)
        groups = self._getQueueGroups(name)
        pending = sum(group.get("pending", 0) for group in groups)
        return {
            "name": name,
            "groups": groups,
            # "length": queue.get("length", 0),
            "length": length,
            "pending": pending,
        }

    def createQueue(
            self, name, group="default", consumeID="0", force=False, existOk=False
    ):
        logger.debug(f"Create Queue {name}-{group}")
        if force:
            self.deleteQueue(name)
        return self._createQueue(
            name, group=group, consumeID=consumeID, existOk=existOk
        )

    def _createQueue(self, name, group="default", consumeID="0", existOk=False):
        try:
            return self.client.xgroup_create(name, group, id=consumeID, mkstream=True)
        except redis.RedisError as e:
            # tracebackInfo = traceback.format_exc()
            logger.warn(f"Redis create queue error: {e}")
            if not existOk:
                raise error.MQError(f"Redis queue {name} existed") from e
            return None

    def deleteQueue(self, *names):
        return self.client.delete(*names)

    # def hasQueue(self, name, group="default"):
    #     try:
    #         queue = self.client.xinfo_stream(name)
    #         groups = self.client.xinfo_groups(name)
    #         return any(g["name"].decode() == group for g in groups)
    #     except Exception:
    #         return False

    def sendMessage(
            self, queue, data, messageID="*", maxlen=1000, trimImmediately=False
    ):
        if self._p2p:
            success = data.get('success')
            if success != 'true':
                return

            for target, data4in in self._sendP2PMessages(queue, data):
                logger.debug(f'send p2p message to {target}: {data4in}')
                self.client.xadd(
                    target, data4in, id=messageID, maxlen=maxlen, approximate=(not trimImmediately)
                )
        else:
            return self.client.xadd(
                queue, data, id=messageID, maxlen=maxlen, approximate=(not trimImmediately)
            )

    def sendEvent(self, queue, data, messageID="*", maxlen=1000, trimImmediately=False):
        return self.client.xadd(
            queue, data, id=messageID, maxlen=maxlen, approximate=(not trimImmediately)
        )

    def _getP2PConnections(self):
        from suanpan.graph import graph

        self._outputConnections = OrderedDict()
        node = graph.getNodeById(g.nodeId)
        for port in node.metadata['def'].ports:
            if port.ioType == 'out':
                port_id = port.uuid
                to_dashboard = False
                connections = graph.getOutputConnections(g.nodeId, port_id)
                for tgt in connections:
                    tgt_node = graph.getNodeById(tgt.process)
                    if tgt_node.metadata['def'].type in [22, 10002]:
                        to_dashboard = True
                        break

                self._outputConnections[port_id] = (to_dashboard, connections)

    def _sendP2PMessages(self, queue, data):
        send_master = False
        for port_id, conns in self._outputConnections.items():
            to_dashboard, connections = conns
            send_master = send_master or to_dashboard
            if not to_dashboard:
                for tgt in connections:
                    d = data.get(port_id)
                    if d:
                        data4in = {"id": data['request_id'], tgt.port: d}
                        extra = data.get('extra')
                        if extra:
                            data4in["extra"] = extra
                        target_queue = f'{queue}-{tgt.process}'
                        yield target_queue, data4in

                data.pop(port_id, None)

        if send_master:
            # connection to dashboard, data must send to master
            yield queue, data

    def recvMessages(
            self,
            queue,
            group="default",
            consumer="unknown",
            noack=False,
            block=False,
            count=1,
            consumeID=">",
    ):
        block = None if not block else 0 if block is True else block
        messages = self.client.xreadgroup(
            group, consumer, {queue: consumeID}, count=count, block=block, noack=noack
        )
        messages = list(self._parseMessagesGenerator(messages, group))

        lostMessageIDs = [
            message["id"]
            for message in messages
            if message["id"] and not message["data"]
        ]
        if lostMessageIDs:
            self.client.xack(queue, group, *lostMessageIDs)
            logger.warning(f"Messages have lost: {lostMessageIDs}")

        return [message for message in messages if message["data"]]

    def subscribeQueue(
            self,
            queue,
            group="default",
            consumer="unknown",
            noack=False,
            block=True,
            count=1,
            consumeID=">",
            delay=0,
            errDelay=1,
            errCallback=logger.error,
    ):
        self.createQueue(queue, group=group, existOk=True)
        logger.debug("Subscribing Messages")
        while True:
            try:
                messages = self.recvMessages(
                    queue,
                    group=group,
                    consumer=consumer,
                    noack=noack,
                    block=block,
                    count=count,
                    consumeID=consumeID,
                )
            except Exception as e:  # pylint: disable=broad-except
                if self._stop_event.is_set():
                    logger.debug(f"stop event is set, break.")
                    break

                errCallback(e)
                logger.debug(f"Error in receiving messages. Wait {errDelay}s")
                gevent.sleep(errDelay)
                self.createQueue(queue, group=group, existOk=True)
                continue

            if not messages:
                logger.debug(f"Received no messages. Wait {delay}s")
                gevent.sleep(delay)
                continue

            for message in messages:
                yield message
                self.client.xack(queue, group, message["id"])

    def recvEvents(
            self,
            queue,
            block=False,
            count=1,
            consumeID="$",
    ):
        block = None if not block else 0 if block is True else block
        events = self.client.xread({queue: consumeID}, count=count, block=block)
        events = list(self._parseMessagesGenerator(events, None))

        return [event for event in events if event["data"]]

    def subscribeEvent(
            self,
            queue,
            block=True,
            count=1,
            consumeID="$",
            delay=0,
            errDelay=1,
            errCallback=logger.error,
    ):
        logger.debug("Subscribing Event")
        while True:
            try:
                events = self.recvEvents(queue, block=block, count=count, consumeID=consumeID)
            except Exception as e:  # pylint: disable=broad-except
                if self._stop_event.is_set():
                    logger.debug(f"stop event is set, break.")
                    break

                errCallback(e)
                logger.debug(f"Error in receiving events. Wait {errDelay}s")
                gevent.sleep(errDelay)
                continue

            if not events:
                # logger.debug(f"Received no events. Wait {delay}s")
                gevent.sleep(delay)
                continue

            for event in events:
                yield event

    def recvPendingMessagesInfo(
            self, queue, group="default", consumer="unknown", start="-", end="+", count=None
    ):
        return self.client.xpending_range(
            queue, group, start, end, count, consumername=consumer
        )

    def retryPendingMessages(
            self,
            queue,
            group="default",
            consumer="unknown",
            count=100,
            maxTimes=3,
            timeout=1,
            errCallback=logger.error,
            maxlen=1000,
            trimImmediately=False,
    ):
        logger.debug("Retrying Pending Messages")

        def _getPendingMessages():
            try:
                return self.recvMessages(
                    queue,
                    group=group,
                    consumer=consumer,
                    block=False,
                    count=count,
                    consumeID="0",
                )
            except Exception:  # pylint: disable=broad-except
                logger.warning("Error in getting pending messages:")
                logger.warning(traceback.format_exc())
                return []

        pendingMessages = {msg["id"]: msg for msg in _getPendingMessages()}
        if not pendingMessages:
            logger.debug("Nothing to retry!")
            return

        pendingInfos = {
            msg["message_id"]: msg
            for msg in self.recvPendingMessagesInfo(
                queue, group=group, consumer=consumer, count=count
            )
        }

        for mid in pendingMessages.keys():
            message = pendingMessages[mid]
            info = pendingInfos.get(mid, {})
            message = utils.merge(message, info)
            data = message["data"]
            deliveredTimes = int(data.get(MQ_DELIVER_KEY, 1))
            if deliveredTimes >= maxTimes:
                logger.error(
                    f"Message {mid} retry failed {deliveredTimes} times. Drop!"
                )
                errCallback(message)
                self.client.xack(queue, group, message["id"])
                continue
            timeSinceDelivered = message.get("time_since_delivered", 0)
            if timeSinceDelivered < timeout:
                logger.warning(
                    f"Message {mid} maybe lost: {timeSinceDelivered} < {timeout}"
                )
                continue
            success = self.client.xack(queue, group, message["id"])
            if success:
                data.update({MQ_DELIVER_KEY: deliveredTimes + 1})
                newMID = self.sendMessage(
                    queue, data, maxlen=maxlen, trimImmediately=trimImmediately
                )
                logger.warning(f"Message send back to queue: {mid} -> {newMID}")

    def _parseMessagesGenerator(self, messages, group):
        for message in messages:
            queue, items = message
            for item in items:
                mid, data = item
                yield {"id": mid, "data": data, "queue": queue, "group": group}

    def groupsInfo(self, key):
        return self.client.xinfo_groups(key)

    def streamInfo(self, key):
        return self.client.xinfo_stream(key)



# coding=utf-8
from __future__ import absolute_import, print_function

from suanpan.utils import json


def loadjson(obj, *args, **kwargs):  # pylint: disable=unused-argument
    import numpy as np
    return np.frombuffer(obj["data"], obj["dtype"]).reshape(obj["shape"])


def loadjsons(s, *args, **kwargs):
    obj = json.loads(s, *args, **kwargs)
    obj["data"] = obj["data"].encode()
    return loadjson(obj)


def loadjsonf(file, *args, **kwargs):
    obj = json.load(file, *args, **kwargs)
    obj["data"] = obj["data"].encode()
    return loadjson(obj)


def dumpjson(npy, *args, **kwargs):  # pylint: disable=unused-argument
    return {"dtype": str(npy.dtype), "shape": npy.shape, "data": npy.tobytes()}


def dumpjsons(npy, *args, **kwargs):
    obj = dumpjson(npy)
    obj["data"] = obj["data"].decode()
    return json.dumps(obj, *args, **kwargs)


def dumpjsonf(npy, fp, *args, **kwargs):
    obj = dumpjson(npy)
    obj["data"] = obj["data"].decode()
    return json.dump(obj, fp, *args, **kwargs)


def load(file, *args, **kwargs):
    import numpy as np
    kwargs.setdefault("allow_pickle", True)
    kwargs.setdefault("fix_imports", True)
    return np.load(file, *args, **kwargs)


def dump(npy, file, *args, **kwargs):
    import numpy as np
    kwargs.setdefault("allow_pickle", True)
    kwargs.setdefault("fix_imports", True)
    return np.save(file, npy, *args, **kwargs)



# coding=utf-8
from __future__ import absolute_import, print_function

from suanpan.app import app
from suanpan.app.modules.base import Module

module = Module()


@module.fn
def init(path):
    app.sio.setStatic(path)



# coding=utf-8
from __future__ import absolute_import, print_function

from suanpan.objects import Context
from suanpan.utils import env

g = Context(
    # App
    debug=env.lazyget("SP_DEBUG", type=env.Bool, default=False),
    port=env.lazyget("SP_PSTREAM_PORT", type=env.Int, default=8888),
    timezone=env.lazyget("SP_TIMEZONE", default="UTC"),
    appType=env.lazyget("SP_APP_TYPE", required=True),
    appParams=env.lazyget("SP_PARAM", default=""),
    nodeInfo=env.lazyget("SP_NODE_INFO", default=""),
    userId=env.lazyget("SP_USER_ID", required=True),
    appId=env.lazyget("SP_APP_ID", required=True),
    nodeId=env.lazyget("SP_NODE_ID", required=True),
    nodeGroup=env.lazyget("SP_NODE_GROUP", default="default"),
    # Api
    host=env.lazyget("SP_HOST", required=True),
    hostTls=env.lazyget("SP_HOST_TLS", type=env.Bool, default=False),
    apiHost=env.lazyget("SP_API_HOST", required=True),
    apiHostTls=env.lazyget("SP_API_HOST_TLS", type=env.Bool, default=False),
    affinity=env.lazyget("SP_AFFINITY"),
    accessKey=env.lazyget("SP_ACCESS_KEY", required=True),
    accessSecret=env.lazyget("SP_ACCESS_SECRET", required=True),
    userIdHeaderField=env.lazyget("SP_USER_ID_HEADER_FIELD", default="x-sp-user-id"),
    userSignatureHeaderField=env.lazyget(
        "SP_USER_SIGNATURE_HEADER_FIELD", default="x-sp-signature"
    ),
    userSignVersionHeaderField=env.lazyget(
        "SP_USER_SIGN_VERSION_HEADER_FIELD", default="x-sp-sign-version"
    ),
    # Screenshots
    screenshotsType=env.lazyget("SP_SCREENSHOTS_TYPE", default="index"),
    screenshotsPattern=env.lazyget("SP_SCREENSHOTS_PATTERN"),
    screenshotsStorageKey=env.lazyget("SP_SCREENSHOTS_STORAGE_KEY"),
    screenshotsThumbnailStorageKey=env.lazyget("SP_SCREENSHOTS_THUMBNAIL_STORAGE_KEY"),
    # Logkit
    logkitUri=env.lazyget("SP_LOGKIT_URI", default=""),
    logkitNamespace=env.lazyget("SP_LOGKIT_NAMESPACE", default="/logkit"),
    logkitPath=env.lazyget("SP_LOGKIT_PATH", default=""),
    logkitEventsAppend=env.lazyget("SP_LOGKIT_EVENTS_APPEND", default="append"),
    logkitLogsLevel=env.lazyget("SP_LOGKIT_LOGS_LEVEL", default="warning"),
    # terminate stream
    termPort=env.lazyget("SP_TERM_PORT", type=env.Int, default=8002),
    currentOS=env.lazyget("SP_OS", default="kubernetes"),
    desktopHome=env.lazyget("SP_DESKTOP_HOME", default=""),
    backendPort=env.lazyget("SP_PORT", type=env.Int, default=7000),
    portStart=env.lazyget("SP_PORT_START", type=env.Int, default=10000),
    portEnd=env.lazyget("SP_PORT_END", type=env.Int, default=20000),
    # service disable
    serviceSioDisable=env.lazyget("SP_SERVICE_SIO_DISABLE", type=env.Bool, default=False),
    serviceHttpDisable=env.lazyget("SP_SERVICE_HTTP_DISABLE", type=env.Bool, default=False),
    serviceErgoEnable=env.lazyget("SP_SERVICE_ERGO_ENABLE", type=env.Bool, default=False),
)



# coding=utf-8
from __future__ import absolute_import, print_function

from suanpan.app.modules.base import Module, Modules
from suanpan.app.modules.core import module as core
from suanpan.app.modules.front import module as front
from suanpan.app.modules.images import module as images
from suanpan.app.modules.node import module as node
from suanpan.app.modules.params import module as params
from suanpan.app.modules.storage import module as storage

modules = Modules()
modules.register("core", core)
modules.register("node", node)
modules.register("front", front)
modules.register("params", params)
modules.register("images", images, enable=False)
modules.register("storage", storage, enable=False)



# coding=utf-8
from __future__ import absolute_import, print_function

from suanpan import error
from suanpan.utils import json


class Variable(object):
    def __init__(self, vars, mstorage, key):
        self.vars = vars
        self.mstorage = mstorage
        self.key = key
        self._value = None

    @property
    def value(self):
        if self._value is None:
            self._value = self.get()
        return self._value

    def get(self, default=None):
        value = self.mstorage.get(self.key)
        value = default if value is None else self.decode(value)
        self._value = value
        return self._value

    def set(self, value, expire=None):
        self._value = value
        self.mstorage.set(
            self.key, self.encode(self._value), expire=expire or self.vars.VARS_EXPIRE
        )

    def delete(self):
        self.mstorage.delete(self.key)

    @classmethod
    def encode(cls, value):
        return value

    @classmethod
    def decode(cls, value):
        return value


class String(Variable):
    pass


class Int(Variable):
    @classmethod
    def encode(cls, value):
        return str(value)

    @classmethod
    def decode(cls, value):
        return int(value)


class Float(Variable):
    @classmethod
    def encode(cls, value):
        return str(value)

    @classmethod
    def decode(cls, value):
        return float(value)


class Json(Variable):
    @classmethod
    def encode(cls, value):
        return json.dumps(value)

    @classmethod
    def decode(cls, value):
        return json.loads(value)


class Variables(object):
    VARS_EXPIRE = None
    VARS_REGISTER_KEY = "global.vars"
    DEFAULT_VAR_TYPE = "string"
    VARS = {
        "string": String,
        "int": Int,
        "float": Float,
        "json": Json,
    }

    def __init__(self, mstorage):
        self.mstorage = mstorage

    def _getVar(self, key, type=DEFAULT_VAR_TYPE):
        var = self.VARS.get(type)
        if not var:
            raise error.MStorageUnknownVarTypeError(type)
        return var(self, self.mstorage, key)

    def _getFullKey(self, key):
        return key

    def _getOriginKey(self, fullKey):
        return fullKey

    def _getRelativeKey(self, key, prefix=None, delimiter="."):
        if not prefix or not key.startswith(prefix):
            return key
        return key[len(prefix) :].lstrip(delimiter)

    def _getRegisterKey(self):
        return self.VARS_REGISTER_KEY

    def _register(self, key, type):
        return self.mstorage.mset(
            self._getRegisterKey(), {key: type}, expire=self.VARS_EXPIRE
        )

    def _getAllRegistered(self):
        return self.mstorage.mget(self._getRegisterKey())

    def _getRegisteredType(self, key):
        return self._getAllRegistered().get(key)

    def _getAllVars(self):
        return [
            self.var(key, type=type) for key, type in self._getAllRegistered().items()
        ]

    def _getVars(self, prefix=None):
        return [
            var
            for var in self._getAllVars()
            if not prefix or var.key.startswith(prefix)
        ]

    def var(self, key, type=None):
        _type = self._getRegisteredType(key)
        if _type is not None and type is not None and _type != type:
            raise error.MStorageVarHasRegisteredAsAnotherTypeError(
                f"{key}({type}) is not {_type}"
            )
        type = type or _type or self.DEFAULT_VAR_TYPE
        var = self._getVar(self._getFullKey(key), type=type)
        self._register(key, type)
        return var

    def vars(self, prefix=None):
        return self._getVars(prefix=prefix)

    def get(self, key):
        type = self._getRegisteredType(key)
        if not type:
            raise error.MStorageUnregisterVarKeyError(key)
        return self.var(key, type=type).get()

    def getall(self, prefix=None):
        return {
            self._getRelativeKey(self._getOriginKey(arg.key), prefix=prefix): arg.get()
            for arg in self._getVars(prefix=prefix)
        }

    def Int(self, key):
        return self.var(key, type="int")

    def Float(self, key):
        return self.var(key, type="float")

    def String(self, key):
        return self.var(key, type="string")

    def Json(self, key):
        return self.var(key, type="json")



#!/usr/bin/env python

"""
suanpan
"""

import itertools
import os
import re
from setuptools import find_packages, setup

VERSION_PARRTERN = r"__version__ = \"([\d\w\.]*)\""
VERSION_FILE = os.path.join("suanpan", "__init__.py")
VERSION = re.findall(VERSION_PARRTERN, open(VERSION_FILE, "r").read())[0]

BASE_REQUIRES = [
    "six>=1.13.0",
    "imageio>=2.4.1",
    "numpy>=1.16.2",
    # "opencv-python-headless>=4.1.2",
    "pandas>=1.3.4",
    "xlrd>=2.0.1",
    "tqdm>=4.28.1",
    "retrying>=1.3.3",
    "tabulate>=0.8.3",
    "colorama>=0.4.1",
    "lostc>=0.1.0",
    "addict>=2.2.0",
    "timeout-decorator>=0.4.1",
    "gevent>=20.5.0,<=23.7.0",
    "python-socketio[client]==5.4.0",
    "gevent-websocket>=0.10.1",
    # 'wsaccel>=0.6.2; platform_system!="Windows"',
    "ujson>=5.8.0",
    "pytz>=2020.1",
]
OSS2_REQUIRES = ["oss2>=2.9.1"]
MINIO_REQUIRES = ["minio==7.0.0"]
REDIS_REQUIRES = [
    "redis>=3.2.0",
    "hiredis>=2.2.3"
]
MYSQL_REQUIRES = ["mysql-connector-python>=8.0.16"]
POSTGRES_REQUIRES = ["psycopg2-binary>=2.8.2"]
HIVE_REQUIRES = [
    'sasl>=0.2.1; platform_system!="Windows"',
    "thrift-sasl>=0.3.0",
    "thrift>=0.11.0",
    "pyhive[hive]>=0.6.1",
]
ODPS_REQUIRES = ["pyodps>=0.8.0"]
EXCEL_REQUIRES = [
    "xlwt>=1.3.0",
    "openpyxl>=3.0.2",
]
ZMQ_REQUIRES = ["pyzmq>=18.1.0"]
INSTALL_REQUIRES = list(
    itertools.chain(
        BASE_REQUIRES,
        OSS2_REQUIRES,
        MINIO_REQUIRES,
        REDIS_REQUIRES,
        ZMQ_REQUIRES,
    )
)
EXTRAS_REQUIRES = {
    'table': list(itertools.chain(
        MYSQL_REQUIRES,
        POSTGRES_REQUIRES,
        HIVE_REQUIRES,
        ODPS_REQUIRES,
    ))
}
README = "README.md"


def read_file(path):
    with open(path, "r") as f:
        return f.read()


fix_packages = ["__suanpan__"]
packages = find_packages(exclude=["tests"])
packages.extend(fix_packages)

setup(
    name="suanpan-core",
    version=VERSION,
    packages=packages,
    package_data={"suanpan": ["helper/example/*/*", "stream/*.so"]},
    license="See License",
    author="yanqinghao",
    author_email="woshiyanqinghao@gmail.com",
    description="Suanpan SDK",
    long_description=read_file(README),
    long_description_content_type="text/markdown",
    zip_safe=False,
    include_package_data=True,
    platforms="any",
    python_requires=">=3.7",
    install_requires=INSTALL_REQUIRES,
    extras_require=EXTRAS_REQUIRES,
    entry_points={
        'console_scripts': [
            'suanpan = suanpan:helper',
        ],
    },
    classifiers=[
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)



# coding=utf-8
from __future__ import absolute_import, print_function

import datetime
import logging
import logging.handlers
import os
import sys
import platform
import pytz
import suanpan
from urllib.parse import urljoin
from suanpan import api, g, path
from suanpan.utils import functional


class Formatter(logging.Formatter):
    def formatTime(self, record, datefmt=None):
        dt = datetime.datetime.fromtimestamp(record.created)
        try:
            tz = pytz.timezone(self.tzName)
        except Exception:  # pylint: disable=broad-except
            tz = pytz.utc
        converted = dt.astimezone(tz)
        return converted.strftime(datefmt) if datefmt else converted.isoformat()

    def __init__(
            self,
            fmt="%(asctime)s :: %(levelname)-10s :: %(message)s",
            datefmt=None,
            timezone="UTC",
    ):
        super().__init__(fmt=fmt, datefmt=datefmt)
        self.tzName = timezone


class Logger(logging.Logger):
    FORMATTER = Formatter(timezone=g.timezone)
    STREAM_LOG_LEVEL = logging.DEBUG
    FILE_LOG_LEVEL = logging.DEBUG
    LOGKIT_LOG_LEVEL = logging.DEBUG
    LOG_FILE_MAX_SIZE = 1024 * 1024 * 1024
    BACKUP_COUNT = 5
    LOG_PATH = "logs"
    LOG_FILE = None
    LOGKIT_URI = g.logkitUri
    LOGKIT_NAMESPACE = g.logkitNamespace
    LOGKIT_PATH = g.logkitPath
    LOGKIT_EVENTS_APPEND = g.logkitEventsAppend

    def __init__(self, name="suanpan"):
        super().__init__(name=name)
        logging.raiseExceptions = False
        self.addStreamHandler(level=logging.DEBUG if g.debug else logging.INFO)
        if self.LOGKIT_URI:
            self.addLogkitHandler(level=logging.getLevelName(g.logkitLogsLevel.upper()))

    def addStreamHandler(self, level=STREAM_LOG_LEVEL, formatter=FORMATTER):
        streamHandler = logging.StreamHandler(stream=sys.stdout)
        streamHandler.setLevel(level)
        streamHandler.setFormatter(formatter)
        self.addHandler(streamHandler)
        return streamHandler

    def addFileHandler(
            self,
            level=FILE_LOG_LEVEL,
            formatter=FORMATTER,
            logPath=LOG_PATH,
            logFile=LOG_FILE,
            logFileMaxSize=LOG_FILE_MAX_SIZE,
            backupCount=BACKUP_COUNT,
    ):
        logFile = logFile or f"{self.name}.log"
        logFilePath = os.path.join(logPath, logFile)
        path.mkdirs(logFilePath, parent=True)
        fileHandler = logging.handlers.RotatingFileHandler(
            logFilePath, maxBytes=logFileMaxSize, backupCount=backupCount
        )
        fileHandler.setLevel(level)
        fileHandler.setFormatter(formatter)
        self.addHandler(fileHandler)
        return fileHandler

    def addLogkitHandler(
            self,
            level=LOGKIT_LOG_LEVEL,
            formatter=FORMATTER,
            uri=LOGKIT_URI,
            namespace=LOGKIT_NAMESPACE,
            socketio_path=LOGKIT_PATH,
            event=LOGKIT_EVENTS_APPEND,
    ):
        url = urljoin(uri, namespace)
        logkitHandler = LoggerLogkitHandler(url, namespace, socketio_path, event)
        logkitHandler.setLevel(level)
        logkitHandler.setFormatter(formatter)
        self.addHandler(logkitHandler)
        return logkitHandler

    @functional.onlyonce
    def logDebugInfo(self):
        self.logPythonVersion()
        self.logSdkVersion()

    @functional.onlyonce
    def logSdkVersion(self):
        self.debug("Suanpan SDK (ver: {})".format(suanpan.__version__))

    @functional.onlyonce
    def logPythonVersion(self):
        self.debug("Python (ver: {})".format(platform.python_version()))

    def makeRecord(
            self,
            name,
            level,
            fn,
            lno,
            msg,
            args,
            exc_info,
            func=None,
            extra=None,
            sinfo=None,
    ):
        extra = {"extra": {"app": g.appId, "node": g.nodeId, **(extra or {})}}
        return super().makeRecord(
            name, level, fn, lno, msg, args, exc_info, func, extra, sinfo
        )


class LoggerProxy(object):
    def __init__(self, loggerOrName):
        self.setLogger(loggerOrName)

    def __getattr__(self, key):
        return getattr(self.logger, key)

    def setLogger(self, loggerOrName):
        self.logger = (
            loggerOrName if isinstance(loggerOrName, Logger) else Logger(loggerOrName)
        )


class LoggerLogkitHandler(logging.Handler):
    def __init__(self, url, namespace, socketio_path, event):
        super().__init__()
        self.client = None
        self.url = url
        self.namespace = namespace
        self.socketio_path = socketio_path
        self.event = event

    def makeClient(self):
        return api.sio(self.url, namespaces=self.namespace, socketio_path=self.socketio_path, wait_timeout=3)

    def send(self, msg):
        if not self.client:
            self.client = self.makeClient()
        elif not self.client.connected:
            self.client.disconnect()
            self.client = self.makeClient()
        self.client.emit(self.event, data=msg, namespace=self.namespace)

    def makePickle(self, record):
        app = record.extra.pop("app")
        data = (app,
                {
                    "level": record.levelname,
                    "title": record.message,
                    "data": record.extra,
                    "time": datetime.datetime.now(datetime.timezone.utc).isoformat(),
                })
        return data

    def emit(self, record):
        try:
            msg = self.makePickle(record)
            self.send(msg)
        except Exception:  # pylint: disable=broad-except
            self.handleError(record)

    def close(self):
        """
        Closes the socket.
        """
        self.acquire()
        try:
            if self.client:
                self.client.disconnect()
                self.client = None
            super().close()
        finally:
            self.release()


rootLogger = Logger("suanpan")
logger = LoggerProxy(rootLogger)



# coding=utf-8
from __future__ import absolute_import, print_function

import os

if os.system('python -c "import tensorflow"') == 0:
    import tensorflow  # pylint: disable=import-error

    print("Tensorflow: ", tensorflow.__version__)  # pylint: disable=no-member
else:
    print("Import tensorflow error. Try to reinstall tensorflow.")
    tfversion = os.environ.get("TENSORFLOW_VERSION", "1.12.0")
    os.system(f"pip install tensorflow=={tfversion}")
    import tensorflow  # pylint: disable=import-error

    print("Tensorflow: ", tensorflow.__version__)  # pylint: disable=no-member



from suanpan import g, error
from suanpan.graph import graph


def lookup(port_id: str):
    connections = graph.getOutputConnections(g.nodeId, port_id)
    if len(connections) != 1:
        raise error.GraphError('invalid service connection')

    service = connections[0]
    svc_node = graph.getNodeById(service.process)
    for port in svc_node.metadata['def'].ports:
        if port.uuid == service.port:
            port_info = port.description.zh_CN
            svc = f'app-{g.appId}-{service.process}'
            return f'{svc}:{port_info}' if port_info else svc



import pickle
import pathlib
from suanpan.storage import storage
from suanpan.log import logger


class Settings(object):
    def __init__(self):
        self._data = None
        self._key = None
        self._local = None

    def __setattr__(self, key, value):
        if key in ['_data', '_key', '_local']:
            super(Settings, self).__setattr__(key, value)
        else:
            if self._data is None:
                self._load()

            self._data[key] = value
            self._save()

    def __setitem__(self, key, value):
        if self._data is None:
            self._load()

        self._data[key] = value
        self._save()

    def __getitem__(self, item):
        if self._data is None:
            self._load()

        return self._data.get(item)

    def __getattribute__(self, item):
        try:
            return super(Settings, self).__getattribute__(item)
        except AttributeError:
            if self._data is None:
                self._load()

            return self._data.get(item)

    def _load(self):
        self._key = storage.getKeyInNodeConfigsStore('settings.pickle')
        self._local = storage.getPathInTempStore('settings.pickle')
        pathlib.Path(storage.tempStore).mkdir(parents=True, exist_ok=True)

        try:
            storage.downloadFile(self._key, self._local)
            with open(self._local, 'rb') as f:
                self._data = pickle.load(f)
        except Exception as e:
            logger.debug(f'failed to load settings: {e}')
            self._data = {}

    def _save(self):
        with open(self._local, 'wb') as f:
            pickle.dump(self._data, f)
        storage.uploadFile(self._key, self._local)



# coding=utf-8
from __future__ import absolute_import, print_function

import base64
import copy
import itertools
import re
from collections import namedtuple

from lostc import collection as lcc

from suanpan import error, g
from suanpan.app import arguments
from suanpan.log import logger
from suanpan.objects import HasName
from suanpan.utils import functional, json

Param = namedtuple("Param", ["uuid", "name", "type"])
Port = namedtuple("Port", ["uuid", "name", "type", "subtype", "description"], defaults=[{}])

DEFAULT_DATA_SUBTYPE_ARGS_MAP = {
    "all": arguments.String,
    "number": arguments.IntOrFloat,
    "bool": arguments.Bool,
    "string": arguments.String,
    "array": arguments.Json,
    "dyadicArray": arguments.Json,
    "json": arguments.Json,
    "file": arguments.File,
    "csv": arguments.Csv,
    "npy": arguments.Npy,
    "model": arguments.Model,
    "image": arguments.Image,
    "system": arguments.String,
    # not used now
    "int": arguments.Int,
    "float": arguments.Float,
    "folder": arguments.Folder,
    "data": arguments.Data,
    "excel": arguments.Excel,
    "table": arguments.Table,
    "dataframe": arguments.Csv,
}

DEFAULT_PARAM_TYPE_ARGS_MAP = {
    "string": arguments.String,
}


class BaseNode(HasName):
    NODE_INFO_KEY = "SP_NODE_INFO"
    DEFAULT_NODE_INFO = {"info": {}, "inputs": {}, "outputs": {}, "params": {}}

    def __init__(self, *args, **kwargs):  # pylint: disable=unused-argument
        super().__init__()
        self._node = self.load()
        self._data_subtype_args_map = copy.copy(DEFAULT_DATA_SUBTYPE_ARGS_MAP)
        self._param_type_args_map = copy.copy(DEFAULT_PARAM_TYPE_ARGS_MAP)

    def __getattr__(self, key):
        value = self._getPort(key) if re.match(r"(in|out)\d+", key) else self.get(key)
        return value or self.__getattribute__(key)

    def _getPort(self, key):
        return lcc.first(
            itertools.chain(self.inputs, self.outputs), lambda i: i.uuid == key
        )

    def get(self, key, default=None):
        for _, collection in self._node.items():
            if isinstance(collection, dict):
                if key in collection:
                    return collection[key]
        return default

    @property
    def info(self):
        return self._node["info"]

    @property
    def inputs(self):
        return self._node["inputs"].values()

    @property
    def ins(self):
        return self._node["inputs"].values()

    @property
    def outputs(self):
        return self._node["outputs"].values()

    @property
    def outs(self):
        return self._node["outputs"].values()

    @property
    def params(self):
        return self._node["params"].values()

    def loadFromEnv(self):
        nodeInfoBase64 = g.nodeInfo
        logger.debug(f"{self.NODE_INFO_KEY}(Base64)='{nodeInfoBase64}'")
        nodeInfoString = base64.b64decode(nodeInfoBase64).decode()
        nodeInfo = json.loads(nodeInfoString)
        return nodeInfo

    def formatNodeInfo(self, nodeInfo):
        inputs = {
            name: Port(**port) for name, port in nodeInfo.get("inputs", {}).items()
        }
        outputs = {
            name: Port(**port) for name, port in nodeInfo.get("outputs", {}).items()
        }
        params = {
            name: Param(**param) for name, param in nodeInfo.get("params", {}).items()
        }
        return {"inputs": inputs, "outputs": outputs, "params": params}

    def defaultNodeInfo(self):
        return copy.deepcopy(self.DEFAULT_NODE_INFO)

    def _updateInfo(self, *infos):
        result = self.defaultNodeInfo()
        keys = result.keys()
        for info in infos:
            for key in keys:
                result[key].update(info.get(key, {}))
        return result

    def load(self):
        return self._updateInfo(self.formatNodeInfo(self.loadFromEnv()))

    def setDataSubtypeArgs(self, *args, **kwargs):
        self._data_subtype_args_map.update(*args, **kwargs)

    def setParamTypeArgs(self, *args, **kwargs):
        self._param_type_args_map.update(*args, **kwargs)

    def _portArg(self, port, prefix):
        n = port.uuid[len(prefix):]
        if port.type == "table":
            return arguments.Table(
                f"{prefix}putData{n}",
                table=f"{prefix}putTable{n}",
                partition=f"{prefix}putPartition{n}",
                alias=port.uuid,
            )
        if port.type == "visual":
            return arguments.Visual(port.name, alias=port.uuid)
        if port.type == "data":
            argtype = self._data_subtype_args_map.get(port.subtype)
            if not argtype:
                raise error.NodeAutoLoadUnsupportedDataSubtypeError(port.subtype)
            return argtype(port.name, alias=port.uuid)
        if port.type in self._data_subtype_args_map:
            logger.info(f"Node auto load {prefix} type: {port.type}")
            argtype = self._data_subtype_args_map.get(port.type)
            if not argtype:
                raise error.NodeAutoLoadUnsupportedDataSubtypeError(port.subtype)
            return argtype(port.name, alias=port.uuid)
        logger.warning(f"Node auto load not support {prefix} type: {port.type}")
        return None

    def _inPortArgs(self):
        inPorts = {p.uuid: p for p in self.ins}.values()
        args = (self._portArg(p, prefix="in") for p in inPorts)
        return [arg for arg in args if arg is not None]

    def _outPortArgs(self):
        outPorts = {p.uuid: p for p in self.outs}.values()
        args = (self._portArg(p, prefix="out") for p in outPorts)
        return [arg for arg in args if arg is not None]

    def _paramArg(self, param):
        argtype = self._data_subtype_args_map.get(param.type)
        if not argtype:
            raise error.NodeAutoLoadUnsupportedParamTypeError(param.type)
        return argtype(param.name)

    def _paramArgs(self):
        args = (self._paramArg(p) for p in self.params if p.name.startswith("param"))
        return [arg for arg in args if arg is not None]

    @functional.lazyproperty
    def inargs(self):
        return self._inPortArgs()

    @functional.lazyproperty
    def outargs(self):
        return self._outPortArgs()

    @functional.lazyproperty
    def paramargs(self):
        return self._paramArgs()

    def json(self):
        return {
            "info": self.info,
            "inputs": [i._asdict() for i in self.ins],
            "outputs": [i._asdict() for i in self.outs],
            "params": [i._asdict() for i in self.params],
        }



# coding=utf-8
from __future__ import absolute_import, print_function

from suanpan import g
from suanpan.api import requests
from suanpan.api.hardware import app, suanpan
from suanpan.utils import functional


class Hardware(object):
    def __init__(self, host, tls=False):
        super().__init__()
        self.host = host
        self.request = Request(self.host, tls=tls)

    @functional.lazyproperty
    def suanpan(self):
        return suanpan.Interface(self.request)

    @functional.lazyproperty
    def apps(self):
        return app.Interface(self.request)


class Request(object):
    def __init__(self, host, tls=False):
        super().__init__()
        self.host = host
        self.tls = tls

    def getUrl(self, path):
        protocol = "https" if self.tls else "http"
        return f"{protocol}://{self.host}{path}"

    def get(self, path, *args, **kwargs):
        return requests.get(self.getUrl(path), *args, **kwargs)

    def post(self, path, *args, **kwargs):
        return requests.post(self.getUrl(path), *args, **kwargs)

    def put(self, path, *args, **kwargs):
        return requests.put(self.getUrl(path), *args, **kwargs)

    def delete(self, path, *args, **kwargs):
        return requests.delete(self.getUrl(path), *args, **kwargs)
